@@ -15,7 +15,11 @@
  */
 package net.sf.cglib.proxy;
 
+import java.io.ByteArrayOutputStream;
 import java.io.IOException;
+import java.io.InputStream;
+import java.lang.ref.PhantomReference;
+import java.lang.ref.ReferenceQueue;
 import java.lang.reflect.Field;
 import java.lang.reflect.Method;
 import java.lang.reflect.Modifier;
@@ -27,6 +31,7 @@ import junit.framework.Test;
 import junit.framework.TestCase;
 import junit.framework.TestSuite;
 import net.sf.cglib.CodeGenTestCase;
+import net.sf.cglib.core.AbstractClassGenerator;
 import net.sf.cglib.core.DefaultNamingPolicy;
 import net.sf.cglib.core.ReflectUtils;
 import net.sf.cglib.reflect.FastClass;
@@ -131,16 +136,36 @@ public class TestEnhancer extends CodeGenTestCase {
         
     }
 
+    public void testFinalizeNotProxied() throws Throwable {
+        Source source = (Source) Enhancer.create(
+                Source.class,
+                null, TEST_INTERCEPTOR);
+
+        try {
+            Method finalize = source.getClass().getDeclaredMethod("finalize");
+            assertNull("CGLIB should enhanced object should not declare finalize() method so proxy objects are not eligible for finalization, thus faster", finalize);
+        } catch(NoSuchMethodException e) {
+            // expected
+        }
+    }
+
     public void testEnhanceObject() throws Throwable {
         EA obj = new EA();
         EA save = obj;
         obj.setName("herby");
         EA proxy = (EA)Enhancer.create( EA.class,  new DelegateInterceptor(save) );
      
-        assertTrue(proxy.getName().equals("herby"));
+        assertEquals("proxy.getName()", "herby", proxy.getName());
 
         Factory factory = (Factory)proxy;
-        assertTrue(((EA)factory.newInstance(factory.getCallbacks())).getName().equals("herby"));
+        assertEquals("((EA)factory.newInstance(factory.getCallbacks())).getName()", "herby", ((EA)factory.newInstance(factory.getCallbacks())).getName());
+
+        Enhancer e = new Enhancer();
+        e.setSuperclass(EA.class);
+        e.setCallbackType(MethodInterceptor.class);
+        Factory factory1 = e.createFactory();
+        assertEquals("((EA)e.createFactory().newInstance(factory.getCallbacks())).getName()",
+                "herby", ((EA)factory1.newInstance(new DelegateInterceptor(save))).getName());
     }
 
     class DelegateInterceptor implements MethodInterceptor {
@@ -243,6 +268,139 @@ public class TestEnhancer extends CodeGenTestCase {
         
     }
 
+    /**
+     * Verifies that the cache in {@link AbstractClassGenerator} SOURCE doesn't
+     * leak class definitions of classloaders that are no longer used.
+     */
+    public void testSourceCleanAfterClassLoaderDispose() throws Throwable {
+        ClassLoader custom = new ClassLoader(this.getClass().getClassLoader()) {
+
+            @Override
+            public Class<?> loadClass(String name) throws ClassNotFoundException {
+                if (EA.class.getName().equals(name)) {
+                    InputStream classStream = this.getClass().getResourceAsStream("/net/sf/cglib/proxy/EA.class");
+                    byte[] classBytes;
+                    try {
+                        classBytes = toByteArray(classStream);
+                        return this.defineClass(null, classBytes, 0, classBytes.length);
+                    } catch (IOException e) {
+                        return super.loadClass(name);
+                    }
+                } else {
+                    return super.loadClass(name);
+                }
+            }
+        };
+
+        PhantomReference<ClassLoader> clRef = new PhantomReference<ClassLoader>(custom,
+                new ReferenceQueue<ClassLoader>());
+
+        buildAdvised(custom);
+        custom = null;
+
+        for (int i = 0; i < 10; ++i) {
+            System.gc();
+            Thread.sleep(100);
+            if (clRef.isEnqueued()) {
+                break;
+            }
+        }
+        assertTrue("CGLIB should allow classloaders to be evicted. PhantomReference<ClassLoader> was not cleared after 10 gc cycles," +
+                "thus it is likely some cache is preventing the class loader to be garbage collected", clRef.isEnqueued());
+
+    }
+
+    protected Object buildAdvised(ClassLoader custom)
+            throws InstantiationException, IllegalAccessException, ClassNotFoundException {
+        final Class<?> eaClassFromCustomClassloader = custom.loadClass(EA.class.getName());
+
+        CallbackFilter callbackFilter = new CallbackFilter() {
+            Object advised = eaClassFromCustomClassloader.newInstance();
+
+            public int accept(Method method) {
+                return 0;
+            }
+
+        };
+
+        // Need to test both orders: "null classloader first" and "null last" just in case
+        Object source_ = enhance(eaClassFromCustomClassloader, null, callbackFilter, TEST_INTERCEPTOR, null);
+        Object source = enhance(eaClassFromCustomClassloader, null, callbackFilter, TEST_INTERCEPTOR, custom);
+        assertSame("Same proxy class is expected since Enhancer with null (default) ClassLoader should use " +
+                        " target class.getClassLoader(), thus the same cache key instance should be reused",
+                source.getClass(), source_.getClass()
+        );
+
+        Object source2 = enhance(eaClassFromCustomClassloader, null, callbackFilter, TEST_INTERCEPTOR, custom);
+        assertSame("enhance should return cached Enhancer when calling with same parameters",
+                source.getClass(), source2.getClass()
+        );
+
+        Object source2_ = enhance(eaClassFromCustomClassloader, null, callbackFilter, TEST_INTERCEPTOR, null);
+        assertSame("Same proxy class is expected since Enhancer with null (default) ClassLoader should use " +
+                " target class.getClassLoader(), thus the same cache key instance should be reused",
+                source.getClass(), source2_.getClass()
+        );
+
+        Object source3 = enhance(eaClassFromCustomClassloader, null, null, TEST_INTERCEPTOR, custom);
+        assertNotSame("enhance should return different instance when callbackFilter differs",
+                source.getClass(), source3.getClass()
+        );
+
+        return source;
+    }
+
+    private static byte[] toByteArray(InputStream input) throws IOException {
+        ByteArrayOutputStream output = new ByteArrayOutputStream();
+        byte[] buffer = new byte[4096];
+        int n = 0;
+
+        while (-1 != (n = input.read(buffer))) {
+            output.write(buffer, 0, n);
+        }
+
+        return output.toByteArray();
+
+    }
+
+    private static class TestFilter implements CallbackFilter {
+        private final int pk;
+
+        TestFilter(int pk) {
+            this.pk = pk;
+        }
+
+        public int accept(Method method) {
+            return 0;
+        }
+
+        @Override
+        public int hashCode() {
+            return 1; // Make sure Enhancer uses equals, not just hashCode alone
+        }
+
+        @Override
+        public boolean equals(Object obj) {
+            return obj instanceof TestFilter && ((TestFilter) obj).pk == pk;
+        }
+    }
+
+    public void testCallbackFilterEqualsVsClassReuse() {
+        Callback[] callbacks = new Callback[]{NoOp.INSTANCE};
+        Object a = Enhancer.create(Source.class, null, new TestFilter(1), callbacks);
+        Object b = Enhancer.create(Source.class, null, new TestFilter(1), callbacks);
+        assertSame("Using the same as per .equal() CallbackFilter, thus Enhancer should reuse the same proxy class",
+                a.getClass(), b.getClass());
+    }
+
+    public void testCallbackFilterNotEqualsVsClassReuse() {
+        Callback[] callbacks = new Callback[]{NoOp.INSTANCE};
+        Object a = Enhancer.create(Source.class, null, new TestFilter(1), callbacks);
+        Object b = Enhancer.create(Source.class, null, new TestFilter(2), callbacks);
+        assertNotSame("Using the different CallbackFilter instances, thus Enhancer should generate new proxy class",
+                a.getClass(), b.getClass());
+    }
+
     public void testRuntimException()throws Throwable{
     
         Source source =  (Source)Enhancer.create(
@@ -488,6 +646,16 @@ public class TestEnhancer extends CodeGenTestCase {
         return e.create();
     }
 
+    public static Object enhance(Class cls, Class interfaces[], CallbackFilter callbackFilter, Callback callback, ClassLoader loader) {
+        Enhancer e = new Enhancer();
+        e.setSuperclass(cls);
+        e.setInterfaces(interfaces);
+        e.setCallbackFilter(callbackFilter);
+        e.setCallback(callback);
+        e.setClassLoader(loader);
+        return e.create();
+    }
+
     public interface PublicClone extends Cloneable {
         Object clone() throws CloneNotSupportedException;
     }
@@ -555,7 +723,7 @@ public class TestEnhancer extends CodeGenTestCase {
         Object instance = enhancer.create();
 
         assertTrue(instance instanceof ClassOnlyX);
-        assertTrue(instance.getClass().equals(type));
+        assertEquals("types of enhancer.createClass() and enhancer.create().getClass() should match", type, instance.getClass());
     }
 
      public void testSql() {
